/*
	Copyright 2000,2001, Olaf van Es.   All Rights Reserved.
*/

#include <OS.h>
#include <USB.h>
#include "ppu.h"
#include "USBKit.h"
#include <InterfaceDefs.h>
#include <stdlib.h>
#include <stdio.h>


#define DEBUG_DRIVER 0
#define ID "PenpUSB"

bool		ppuInputDevice::sActive = false;
int32		ppuInputDevice::sSpeed = 0;
thread_id	ppuInputDevice::sThread = B_ERROR;
bigtime_t 	gDblClickTime;

char version[]="$VER: wacom usb 1.2.4 (24 april 2001)";
char copyright[]="Wacom PenPartner/Graphire/Intuos USB driver © 2000,2001 by Olaf van Es <olaf.van.es@zonnet.nl>\n";


BInputServerDevice*
instantiate_input_device()
{
	// this is where it all starts
	// make sure this function is exported!

	return (new ppuInputDevice());
}


ppuInputDevice::ppuInputDevice()
	: BInputServerDevice()
{
	// register your device(s) with the input_server
	// the last device in the list should be NULL

	input_device_ref	penpartnerusbDevice			= { "Wacom USB",
													B_POINTING_DEVICE,
													NULL };
	input_device_ref	*ppuDeviceList[2]	= {&penpartnerusbDevice, NULL};

	RegisterDevices(ppuDeviceList);	
}


ppuInputDevice::~ppuInputDevice()
{
	// cleanup
}


status_t
ppuInputDevice::InitCheck()
{
	// do any init code that could fail here
	// you will be unloaded if you return false

	return (BInputServerDevice::InitCheck());
}


status_t
ppuInputDevice::SystemShuttingDown()
{
	// do any cleanup (ie. saving a settings file) when the
	// system is about to shut down

	return (BInputServerDevice::SystemShuttingDown());
}


status_t
ppuInputDevice::Start(
	const char	*device,
	void		*cookie)
{
	// start generating events
	// this is a hook function, it is called for you
	// (you should not call it yourself)

	sActive = true;
//	get_mouse_speed(&sSpeed);
	sThread = spawn_thread(penpumain, device, B_DISPLAY_PRIORITY, this);

	resume_thread(sThread);

	return (B_NO_ERROR);
}


status_t
ppuInputDevice::Stop(
	const char	*device,
	void		*cookie)
{
	// stop generating events
	// this is a hook function, it is called for you
	// (you should not call it yourself)

	status_t err = B_OK;

	sActive = false;
	wait_for_thread(sThread, &err);
	sThread = B_ERROR;

	return (B_NO_ERROR);
}


status_t
ppuInputDevice::Control(
	const char	*device,
	void		*cookie,
	uint32		code,
	BMessage	*message)
{
	// respond to changes in the system
	// this is a hook function, it is called for you
	// (you should not call it yourself)

	switch (code) {
		case B_CLICK_SPEED_CHANGED:
			if (B_OK != get_click_speed(&gDblClickTime)) gDblClickTime = 250000;
			break;
			}


	return (B_NO_ERROR);
}


int32
ppuInputDevice::penpumain(
	void	*arg)
{

	uint16 x = 0, y = 0;
	uint16 pres = 0;
	int8 prespen = 0;
	int32 mbuttons;
	int8 mousewheel = 0;
	bool ismouse = 0;
	bool leftmb = 0, rightmb = 0, middlemb = 0, eraser = 0; 
	bool vorig1 = 0, vorig2 = 0, vorig3 = 0;
	bool nabij= false;
	bool penpartner = 0, graphire = 0, intuos = 0, pl500 = 0;
	float xval = 0.0f, yval = 0.0f, pressure = 0.0f;
	float xval1 = 0.0f, yval1 = 0.0f;
	float xval2 = 0.0f, yval2 = 0.0f;
	char buf[30];
	int i,j,hub;
	int databytes = 10; //default for intuos
	int8 tiltx = 0, tilty = 0;
	float ftiltx = 0.0f, ftilty = 0.0f;

	int32		clickcount = 0;
	bigtime_t	lastclick = 0;
	float 	lastClickedX = -1.0f, lastClickedY = -1.0f;
	float	deltaX = 0.0f, deltaY = 0.0f;

	float maxx = 1.0f, maxy = 1.0f;
	
	ppuInputDevice *penpartnerusbDevice = (ppuInputDevice *)arg;

//	USBDevice dev("/dev/bus/usb/0/1");
	uchar data[12];

// yes this is not a really nice way to find the device you want.

for (hub=0;hub<2;hub++){
	for (j=0;j<10;j++){
		for (i=0;i<10;i++){
		if (hub==0) {
			sprintf(buf,"/dev/bus/usb/%i/%i",j,i);
			}
		else 
			{
			sprintf(buf,"/dev/bus/usb/%i/%i/%i",hub-1,j,i);
			}
		USBDevice dev(buf);
//		if ((dev.VendorID()==0x56a)&&(dev.ProductID()==0x0)){

		if (dev.VendorID()==0x56a){ //wacom
			if (dev.ProductID() == 0x0){
				penpartner=true;
				maxx=5040.0f;
				maxy=3780.0f;
				databytes=7;
			}
			if (dev.ProductID() == 0x10){
				graphire=true;
				maxx=10206.0f;
				maxy=7422.0f;
				databytes=8;
				}
			if (dev.ProductID() == 0x20) {
				intuos=true;    // 4x5"
				maxx=12700.0f;
				maxy=10360.0f;
				}
			if (dev.ProductID() == 0x21) {
				intuos=true;    // 6x8"
				maxx=20320.0f;
				maxy=15040.0f;
				}
			if (dev.ProductID() == 0x22) {
				intuos=true;    // 9x12"
				maxx=30480.0f;
				maxy=23060.0f;
				}
			if (dev.ProductID() == 0x23) {
				intuos=true;    // 12x12"
				maxx=30480.0f;
				maxy=30480.0f;
				}
			if (dev.ProductID() == 0x24) {
				intuos=true;    // 12x18"
				maxx=47720.0f;
				maxy=23060.0f;
				}
			if (dev.ProductID() == 0x31) {
				pl500=true;    // PL500
				maxx=12328.0f;
				maxy=9256.0f;
				databytes=8;
				}
				
				
			if (!penpartner && !graphire && !intuos && !pl500)
				{
				goto einde;
				}
			
			
			
			if(dev.InitCheck() || dev.SetConfiguration(dev.ConfigurationAt(0))){
				printf("Cannot open and configure: %s\n",buf);
				return 1;
			}
	
			const USBEndpoint *ept = dev.ActiveConfiguration()->InterfaceAt(0)->EndpointAt(0);	
	
			dev.ControlTransfer(0x21,0x0b,1,0,0,0); // you won't believe the amount of time it took to figure this one out.

			if (B_OK != get_click_speed(&gDblClickTime)) gDblClickTime = 250000;



			while (penpartnerusbDevice->sActive) {
			if(ept->InterruptTransfer(data,databytes)==databytes){
			nabij=false;			
			if(penpartner){
				x = data[2] << 8 | data[1];
				y = data[4] << 8 | data[3];

				prespen=data[6];
				eraser = (data[5] & 0x20);
				rightmb = (data[5] & 0x40);
				leftmb = ((prespen > -80) &&!(data[5] & 0x20));
				pressure = (float)(prespen + 120)/240.0f;
				nabij = true;
			}
			if(graphire){
				x = data[3] << 8 | data[2];
				y = data[5] << 8 | data[4];

				pres=data[7] << 8 | data[6];
				if (data[1] & 0x40) {
				  ismouse=true;
				  mousewheel = data[6];
				  }
				eraser = (data[1] & 0x20);
				rightmb = (data[1] & 0x02);
				leftmb = (data[1] & 0x01);
				middlemb = (data[1] & 0x04);
				nabij = (data[1] & 0x80);			
				pressure = (float)(pres)/511.0f;
			}
			if(intuos && (data[0]==0x02) && !(((data[1]>>5)& 0x03)==0x02))
				{
				x = data[2] << 8 | data[3];
				y = data[4] << 8 | data[5];
				pres=data[6] << 2 | ((data[7] >> 6) & 0x03);
//				eraser = (data[1] & 0x20);  eraser is een tool-id
				rightmb = (data[1] & 0x02);
				leftmb = (pres > 180);
				middlemb = (data[1] & 0x04);
				nabij = (data[1] & 0x40);			
				pressure = (float)(pres)/1023.0f;
	// tilt -128 ... 127 maken
				tiltx = ((data[7] & 0x3f) << 2) | ((data[8] & 0x80) >> 6);
				tilty = (data[8] & 0x7f) << 1;
	// floats maken
				ftiltx = (float)tiltx / 128.0f;
				ftilty = (float)tilty / 128.0f; 		
			}
			if (pl500){ //this is never tested
				nabij= ( data[1] & 0x20);
				x = data[2] << 8 | data[3];
				y = data[5] << 8 | data[6];
				pres = (data[4] & 0x04) >> 2 | (data[7] & 0x7f) << 1;
				rightmb = (data[4] & 0x10);
				leftmb = (data[4] & 0x08);				
				middlemb = (data[4] & 0x20);
				pressure = (float)(pres)/511.0f;
			}

			if (pressure > 1) pressure = 1.0f;
			if (pressure < 0) pressure = 0.0f;
			xval = (float)(x)/maxx;
			yval = (float)(y)/maxy; 

// filter some jitter control
			
			xval2 = xval1-xval; if (xval2 < 0) xval2 = -xval2;
			if (xval2 <0.00125) xval = xval1;
			xval1 = xval;

			
			yval2 = yval1-yval; if (yval2 < 0) yval2 = -yval2;
			if (yval2 < 0.001667) yval = yval1;
			yval1 = yval;
//
			mbuttons = (leftmb ? 0x1 : 0x0)|(rightmb ? 0x2 : 0x0)|(middlemb ? 0x4 : 0x0);


// 				bewegen met die muis eh cursor
		if (nabij) {
		
			bigtime_t	now = system_time();

			if (now - lastclick < gDblClickTime) {
				deltaX = xval - lastClickedX; if (deltaX < 0) deltaX = -deltaX;
				deltaY = yval - lastClickedY; if (deltaY < 0) deltaY = -deltaY;
				if ((deltaX < (50.0 / maxx)) && (deltaY < (50.0 / maxy))) {
					xval = lastClickedX; yval = lastClickedY;
				}
			}
					

				BMessage	*event = new BMessage(B_MOUSE_MOVED);
				event->AddInt64("when", now);
				event->AddInt32("buttons", mbuttons);
				event->AddFloat("x", xval);
				event->AddFloat("y", yval);
				event->AddFloat("be:tablet_pressure", pressure);
				event->AddFloat("be:tablet_x", xval);
				event->AddFloat("be:tablet_y", yval);
				event->AddInt32("be:tablet_eraser", eraser);
				if (intuos){
					event->AddFloat("be:tablet_tilt_x", ftiltx);
					event->AddFloat("be:tablet_tilt_y", ftilty);
				}	
				penpartnerusbDevice->EnqueueMessage(event);

// en dan nu de muisknoppen
		
		if ((leftmb && !vorig1)||(rightmb && !vorig2)||(middlemb && !vorig3))
		{
			bigtime_t	now = system_time();
			BMessage	*event = new BMessage(B_MOUSE_DOWN);

			if ((now - lastclick > gDblClickTime) || ((deltaX > (50.0 / maxx)) && (deltaY > (50.0 / maxy)))){
				clickcount = 1; lastClickedX = xval; lastClickedY = yval;
			} else {
				++clickcount; xval = lastClickedX; yval = lastClickedY;
			}
			lastclick = now;
			
			
			event->AddInt64("when", now);
			event->AddInt32("buttons", mbuttons);
			event->AddFloat("x", xval);
			event->AddFloat("y", yval);
			event->AddFloat("be:tablet_pressure", pressure);
			event->AddInt32("clicks", clickcount);
			event->AddInt32("modifiers", modifiers());
			penpartnerusbDevice->EnqueueMessage(event);

		}

		if ((!leftmb && vorig1)||(!rightmb && vorig2)||(!middlemb && vorig3))
		{
			bigtime_t	now = system_time();
			BMessage	*event = new BMessage(B_MOUSE_UP);
			event->AddInt64("when", now);
			event->AddInt32("buttons", mbuttons);
			event->AddFloat("x", xval);
			event->AddFloat("y", yval);
			event->AddInt32("clicks", 0x0);
			event->AddInt32("modifiers", modifiers());	
			penpartnerusbDevice->EnqueueMessage(event);
	  	}
		
		if (ismouse) {
			bigtime_t now = system_time();
			BMessage 	*event = new BMessage(B_MOUSE_WHEEL_CHANGED);
			event->AddInt64("when", now);
			event->AddFloat("be:wheel_delta_y",mousewheel*2);
			penpartnerusbDevice->EnqueueMessage(event);
			}
			  	
				vorig1 = leftmb;
				vorig2 = rightmb;
				vorig3 = middlemb;
				
				snooze(5000);
				} // end if data
				else snooze(100000); // nabij 
				}
				else snooze(5000); // was 1000000 toolid ?
			} // end while
			break;
		} // end if penpartner

		}	// end for i
	} // end for j
} // end for hub
einde:
	return (B_OK);
}

